enum
{
	FUNC_NULL=0,
	FUNC_OPENEX,
	FUNC_OPENDIR,
	FUNC_FOLLOW,
	FUNC_FOCUSNP,
	FUNC_REFRESH,
	FUNC_NEXTSRC,
	FUNC_RESETSRC,
};

const string8 func_name[] =
{
	"Disabled",
	"Open in external viewer",
	"Open directory",
	"Toggle follow cursor",
	"Focus playing",
	"Refresh",
	"Next source",
	"First source",
};

const string8 cfg_src_list_defaults[] =
{
	"$replace(%_path%,%_filename_ext%,)folder.jpg",
	"-match:$replace(%_path%,%_filename_ext%,)*",
	"-components\\default.png",
};

class cfg_src_list_t : public cfg_var
{
private:
	struct entry
	{
		string8 value;
		entry(const char * p_value) : value(p_value) {}
		entry() {}
	};
	ptr_list_t<entry> data;
public:
	~cfg_src_list_t() {data.delete_all();}
	virtual bool get_raw_data(write_config_callback * out)
	{
		unsigned n,m=data.get_count();
		for(n=0;n<m;n++)
		{
			out->write_string(data[n]->value);
		}
		return true;
	}
	virtual void set_raw_data(const void * pdata,int psize)
	{
		data.delete_all();
		read_config_helper r(pdata,psize);
		for(;;)
		{
			entry * item = new entry;
			if (!r.read_string(item->value)) {delete item; break;}
			data.add_item(item);
		}
		on_change();
	}

	virtual void reset()
	{
		data.delete_all();
		unsigned n;
		for(n=0;n<tabsize(cfg_src_list_defaults);n++)
		{
			data.add_item(new entry(cfg_src_list_defaults[n]));
		}
		on_change();
	}

	const char * get_value(unsigned idx) const
	{
		if (idx>=0 && idx<data.get_count()) return data[idx]->value;
		else return 0;
	}

	unsigned find_item(const char * value) const
	{
		unsigned n,m=data.get_count();
		for(n=0;n<m;n++) if (!stricmp_utf8(data[n]->value,value)) return n;
		return (unsigned)(-1);
	}

	unsigned add_item(const char * value)
	{
		on_change();
		return data.add_item(new entry(value));
	}

	void remove_item(unsigned idx)
	{
		data.delete_by_idx(idx);
		on_change();
	}

	void modify_item(unsigned idx,const char * value)
	{
		data[idx]->value = value;
		on_change();
	}

	unsigned get_count() const {return data.get_count();}

	void swap(unsigned idx1,unsigned idx2) {
		data.swap(idx1,idx2);
		on_change();
	}

	cfg_src_list_t(const char * name) : cfg_var(name)
	{
		reset();
		on_change();
	}

	void format_display(unsigned n,string_base & out) const
	{
		out = get_value(n);
	}

};

static cfg_src_list_t cfg_src_list("albumart_srclist");
static cfg_int cfg_centeralbum("albumart_center_image",1);
static cfg_int cfg_expandalbum("albumart_expand_image",0);
static cfg_int cfg_shrinkalbum("albumart_shrink_image",0);
static cfg_int cfg_aspectratio("albumart_maintain_aspect_ratio",1);
static cfg_int cfg_pixelborder("albumart_pixel_border",0);
static cfg_int cfg_selected("albumart_display_selected",0);
static cfg_int cfg_usebg("albumart_use_background",0);
static cfg_int cfg_bgcol("albumart_background_color",0xFFFFFF);
static cfg_int cfg_lftclickfunc("albumart_left_click_function",FUNC_REFRESH);
static cfg_int cfg_mdlclickfunc("albumart_middle_click_function",FUNC_OPENEX);
static cfg_int cfg_dblclickfunc("albumart_double_click_function",FUNC_FOLLOW);
static cfg_int cfg_interpolationmode("albumart_interpolation_mode",3);
static cfg_int cfg_usecontrol("albumart_use_control",0);
static cfg_int cfg_minheight("albumart_min_height",0);
static cfg_int cfg_resetsource("albumart_reset_source",1);
static cfg_int cfg_padding("albumart_padding",0);