#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "resource.h"
#include <commctrl.h>
#include <windowsx.h>

#include "../shared/ui_extension_with_helpers.h"

DECLARE_COMPONENT_VERSION("Album list panel",
						  
						  "0.1.8",
						  
						  "allows you to browse through database\n\n"
						  "based upon albumlist 2.0\n"
						  "compiled: " __DATE__ "\n"
						  "with UI Extension SDK version: " UI_EXTENSION_VERSION
						  
						  );

class NOVTABLE albumart_control_base : public service_base
{
public:
	static GUID get_class_guid()
	{
		// {5549FA34-673C-4c5b-B6CE-5CDC4813E7B4}
		static const GUID guid = 
		{ 0x5549fa34, 0x673c, 0x4c5b, { 0xb6, 0xce, 0x5c, 0xdc, 0x48, 0x13, 0xe7, 0xb4 } };
		return(guid);
	}

	static albumart_control_base* get() 
	{
		return(service_enum_create_t(albumart_control_base,0));
	}
	
	virtual bool is_controllable()=0;
	virtual bool set_image(metadb_handle * track)=0;
	virtual bool set_image()=0;
	virtual void redraw()=0;
	virtual void lock(bool l)=0;
	virtual bool is_locked()=0;
};

bool albumart_focus(metadb_handle * track)
{
	albumart_control_base * foo_uie_albumart = albumart_control_base::get();
	if(!foo_uie_albumart) return false;
	if(!foo_uie_albumart->is_controllable()) return false;
	if(track){
		foo_uie_albumart->lock(true);
		if(foo_uie_albumart->set_image(track))
			foo_uie_albumart->redraw();
	}else{
		foo_uie_albumart->lock(false);
		if(foo_uie_albumart->set_image())
			foo_uie_albumart->redraw();
	}
	foo_uie_albumart->service_release();
	return true;
}

class IDropSource_albumlist : public IDropSource
{
	long refcount;
	HWND wnd;
public:
	IDropSource_albumlist() : refcount(1) {}
	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid,void ** ppvObject)
	{
		if (ppvObject==0) return E_INVALIDARG;
		else if (iid == IID_IUnknown) {AddRef();*ppvObject = (IUnknown*)this;return S_OK;}
		else if (iid == IID_IDropSource) {AddRef();*ppvObject = (IDropSource*)this;return S_OK;}
		else return E_NOINTERFACE;
	}
	virtual ULONG STDMETHODCALLTYPE AddRef() {return InterlockedIncrement(&refcount);}
	virtual ULONG STDMETHODCALLTYPE Release()
	{
		return InterlockedDecrement(&refcount);
	}

	virtual HRESULT STDMETHODCALLTYPE QueryContinueDrag(BOOL fEscapePressed,DWORD grfKeyState)
	{
		if (fEscapePressed || (grfKeyState&MK_RBUTTON)) return DRAGDROP_S_CANCEL;
		else if (!(grfKeyState&MK_LBUTTON))
		{
			if (wnd)
			{
				DWORD pts = GetMessagePos();
				POINT pt = {(short)LOWORD(pts),(short)HIWORD(pts)};
				HWND pwnd = WindowFromPoint(pt);
				if (pwnd==wnd || GetParent(pwnd)==wnd) return DRAGDROP_S_CANCEL;
				else return DRAGDROP_S_DROP;
			}
			else return DRAGDROP_S_CANCEL;
		}
		else return S_OK;
	}

	virtual HRESULT STDMETHODCALLTYPE GiveFeedback(DWORD dwEffect)
	{
		if (wnd)
		{
			DWORD pts = GetMessagePos();
			POINT pt = {(short)LOWORD(pts),(short)HIWORD(pts)};
			HWND pwnd = WindowFromPoint(pt);
			if (pwnd==wnd || GetParent(pwnd)==wnd)
			{
				SetCursor(uLoadCursor(0,IDC_NO));
				return S_OK;
			}
		}
		
		return DRAGDROP_S_USEDEFAULTCURSORS;
	}
	IDropSource_albumlist(HWND wnd_dbe) : wnd(wnd_dbe) {};

};

struct
{
	const char * name, * value;
} cfg_view_list_defaults[] =
{
	{"by album","%album%[ '['$if2(%album artist%,%artist%)']']|[[%disc%.][$num(%tracknumber%,2).] ]$if2(%title%,%_filename_ext%)"},
	{"by artist","artist|[%album%]|[[%disc%.][$num(%tracknumber%,2).] ]$if2(%title%,%_filename_ext%)"},
	{"by genre","genre|[$if2(%album artist%,%artist%)] - [%album%]|[[%disc%.][$num(%tracknumber%,2).] ]$if2(%title%,%_filename_ext%)"},
	{"by year","date|[$if2(%album artist%,%artist%)] - [%album%]|[[%disc%.][$num(%tracknumber%,2).] ]$if2(%title%,%_filename_ext%)"},
	{"by artist/album","[$if2(%album artist%,%artist%) - ]%album%|$if(%album artist%,[[%disc%.][$num(%tracknumber%,2).] ]%artist% - $if2(%title%,%_filename_ext%),[[%disc%.][$num(%tracknumber%,2).] ]$if2(%title%,%_filename_ext%))"},
};

static const char directory_structure_view_name[] = "by directory structure";

class cfg_view_list_t : public cfg_var
{
private:
	struct entry
	{
		string8 name;
		string8 value;
		entry(const char * p_name,const char * p_value) : name(p_name), value(p_value) {}
		entry() {}
	};
	ptr_list_t<entry> data;
public:

	~cfg_view_list_t() {data.delete_all();}
	virtual bool get_raw_data(write_config_callback * out)
	{
		unsigned n,m=data.get_count();
#if 0
		if (m==tabsize(cfg_view_list_defaults))
		{
			bool same = true;
			for(n=0;n<m;n++)
			{
				if (strcmp(data[n]->name,cfg_view_list_defaults[n].name) || strcmp(data[n]->value,cfg_view_list_defaults[n].value)) {same=false; break;}
			}
			if (same) return false;
		}
#endif
		for(n=0;n<m;n++)
		{
			out->write_string(data[n]->name);
			out->write_string(data[n]->value);
		}
		return true;
	}
	virtual void set_raw_data(const void * pdata,int psize)
	{
		data.delete_all();
		read_config_helper r(pdata,psize);
		for(;;)
		{
			entry * item = new entry;
			if (!r.read_string(item->name)) {delete item; break;}
			if (!r.read_string(item->value)) {delete item; break;}
			data.add_item(item);
		}
	}

	virtual void reset()
	{
		data.delete_all();
		unsigned n;
		for(n=0;n<tabsize(cfg_view_list_defaults);n++)
		{
			data.add_item(new entry(cfg_view_list_defaults[n].name,cfg_view_list_defaults[n].value));
		}
	}

	const char * get_name(unsigned idx) const
	{
		if (idx>=0 && idx<data.get_count()) return data[idx]->name;
		else return 0;
	}

	const char * get_value(unsigned idx) const
	{
		if (idx>=0 && idx<data.get_count()) return data[idx]->value;
		else return 0;
	}

	unsigned find_item(const char * name) const
	{
		unsigned n,m=data.get_count();
		for(n=0;n<m;n++) if (!stricmp_utf8(data[n]->name,name)) return n;
		return (unsigned)(-1);
	}

	unsigned add_item(const char * name,const char * value)
	{
		return data.add_item(new entry(name,value));
	}

	void remove_item(unsigned idx)
	{
		data.delete_by_idx(idx);
	}

	void modify_item(unsigned idx,const char * name,const char * value)
	{
		entry * p_entry = data[idx];
		p_entry->name = name;
		p_entry->value = value;
	}

	unsigned get_count() const {return data.get_count();}

	void swap(unsigned idx1,unsigned idx2) {data.swap(idx1,idx2);}

	cfg_view_list_t(const char * name) : cfg_var(name)
	{
		reset();
	}

	void format_display(unsigned n,string_base & out) const
	{
		out = get_name(n);
		out += " : ";
		out += get_value(n);
	}
};

static cfg_view_list_t cfg_view_list("viewlist");



static cfg_string cfg_sort_order("sort_order","$if2(%album artist%,%artist%) - %album%|$num(%disc%,9)|$num(%tracknumber%,9)|$if2(%title%,%_filename_ext%)");
static cfg_string cfg_playlist_name("playlist_name","Database view");

static cfg_int cfg_show_numbers("show_numbers",0),cfg_show_numbers2("show_numbers2",0);
static cfg_int cfg_warning_displayed("warning_displayed",1);
static cfg_int cfg_dblclk("dblclk_to_playlist",0);
static cfg_int cfg_middle("middle_button_action",3);

static cfg_int cfg_use_custom_colours("custom_colours",0);
static cfg_int cfg_colour_back("back_colour",0xFFFFFF);
static cfg_int cfg_colour_text("text_colour",0);
static cfg_int cfg_colour_line("line_colour",0);

static cfg_int cfg_sorttree("sorttree",0);

static cfg_int cfg_autosend("autosend",1);
static cfg_int cfg_keyb("processkeyb",1);

static cfg_int cfg_height("height",0);
static cfg_int cfg_autoplay("autoplay",1);

static cfg_int cfg_populate("populate",1);

static cfg_int cfg_hscroll("hscroll",1);
static cfg_int cfg_frame("window_frame",1);
static cfg_int cfg_postprocess("processing_mode",0);

static cfg_int cfg_resize_left("reszieleft",1);
static cfg_int cfg_show_root("show_root",1);

static cfg_int cfg_use_custom_indent("use_indent",0);
static cfg_int cfg_indent("indentation",19);

static cfg_int cfg_custom_item_height("custom_item_height",0);
static cfg_int cfg_item_height("item_height",4);

inline static uLOGFONT get_def_font()
{
	uLOGFONT foo;
	uGetDefaultFont(&foo);
	return foo;
}

static cfg_struct_t<uLOGFONT> cfg_font("font",get_def_font());


static metadb * g_metadb;

class user_data
{
public:
	HTREEITEM ti;
	menu_manager * p_menu_manager;
	unsigned IDM_MANAGER_BASE;
	ptr_list_autodel_t<string_simple> views;
	user_data() : p_menu_manager(0), ti(0), IDM_MANAGER_BASE(0) {};
};

class album_list_window;

class node
{
private:
	string_simple value;
	node * parent;
	ptr_list_t<node> children;
	metadb_handle_list entries;
	bool b_sorted,b_bydir;
	album_list_window * p_dbe;
	static int sortproc(const node * &n1,const node * &n2)
	{
		return uStringCompare(n1->value,n2->value);
	}
public:

	void sort_children()
	{
		children.sort(sortproc);
		unsigned n;
		for(n=0;n<children.get_count();n++)
			children[n]->sort_children();
	}

	void sort_entries();//for contextmenu

	const metadb_handle_list & get_entries() const {return entries;}

	void create_new_playlist() const
	{
		playlist_switcher * api = playlist_switcher::get();
		if (api)
		{
			metadb_handle_list dummy;
			string8 name = value;
			if (name.is_empty()) name = "All music";
			unsigned idx = api->create_playlist_fixname(name,dummy);
			if (idx != (unsigned)(-1))
			{
				api->set_active_playlist(idx);
				send_to_playlist(true);
			}
		}
	}

	void send_to_playlist(bool replace) const
	{
		playlist_oper::get()->add_items(entries,replace,-1,true,true);
	}

	void remove_from_db() const
	{
		unsigned n;
		for(n=0;n<entries.get_count();n++)
			entries[n]->handle_user_requested_remove();
	}

	void remove_dead() const
	{
		unsigned n;
		for(n=0;n<entries.get_count();n++)
		{
			if (input::is_entry_dead(entries[n]->handle_get_location()))
				entries[n]->handle_entry_dead();
		}
	}

	const char * get_sort_data() const {return value;}
	
	node(node * p_parent,const char * p_value,unsigned p_value_len, album_list_window * dbe) : p_dbe(dbe)
	{
		parent = p_parent;

		if (p_value)
		{
			value.set_string(p_value,p_value_len);
		}
		b_sorted = false;
	}
	inline void set_bydir(bool p) {b_bydir = p;}

	~node()
	{
		entries.delete_all();
		children.delete_all();
	}

	const char * get_val()
	{
		return value.is_empty() ? "All music" : (const char*)value;
	}

	void add_entry(metadb_handle * p_entry)
	{
		entries.add_item(p_entry->handle_duplicate());
	}

	node * add_child(metadb_handle * p_entry,const char * value,unsigned value_len)
	{
		node * found = 0;
		unsigned n,m=children.get_count();
		for(n=0;n<m;n++)
		{
			if (!stricmp_utf8_stringtoblock(children[n]->get_sort_data(),value,value_len)) {found = children[n];break;}
		}
		
		if (found)
		{
			return found;
		}
		else
		{
			node * temp = new node(this,value,value_len,p_dbe);
			children.add_item(temp);
			return temp;
		}
	}

	void reset() {children.delete_all();entries.delete_all();b_sorted=false;}

	inline const ptr_list_t<node> & get_children() {return children;}

	unsigned get_num_children() const
	{
		return children.get_count();
	}

	unsigned get_num_entries() const
	{
		return entries.get_count();
	}

};




static string8 g_formatbuf;

class album_list_window : public container_ui_extension
{
	static const char * class_name;
	bool initialised;
	WNDPROC treeproc;

	bool dragging,clicked;
	DWORD clickpoint;
	int indent_default;

protected:
	static ptr_list_t<album_list_window> list_wnd;
	HWND wnd_tv;
	node * p_selection;

public:

	string8 view;

	bool is_bydir()
	{
		return !stricmp_utf8(view,directory_structure_view_name);
	}

	const char * get_hierarchy()
	{
		unsigned idx = cfg_view_list.find_item(view);
		if (idx!=(unsigned)(-1)) return cfg_view_list.get_value(idx);
		return "N/A";
	}

	node r00t;

	LRESULT WINAPI on_message(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);

	LRESULT WINAPI on_hook(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);
	static LRESULT WINAPI hook_proc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);

	album_list_window();

	void refresh_tree();
	void update_colours();
	void update_item_height();
	static void update_all_colours();
	static void update_all_heights();
	static void update_all_item_heights();
	static void update_all_indents();

	~album_list_window();

	static const GUID extension_guid;

	virtual const GUID & get_extension_guid() const
	{
		return extension_guid;
	}


	virtual void get_name(string_base & out)const;
	virtual void get_category(string_base & out)const;

	virtual void get_config(cfg_var::write_config_callback * out);
	void set_config(cfg_var::read_config_helper * config);

	unsigned get_type () const{return ui_extension_flag::TYPE_PANEL;}

	virtual class_data & get_class_data()const 
	{
		__implement_get_class_data("{606E9CDD-45EE-4c3b-9FD5-49381CEBE8AE}", false);
	}

	static void update_all_window_frames();

	virtual void build_menu(HMENU menu, unsigned base, POINT pt, bool show_shortcuts, void * & user_data); 
	virtual void menu_action(HMENU menu, unsigned base, unsigned id, void * user_data); 

	friend class font_notify;
};

void node::sort_entries()//for contextmenu
{
	if (!b_sorted)
	{
		entries.remove_duplicates_quick();
		
		if (b_bydir) entries.sort_by_format("%_path%|$num(%_subsong%,9)",0);
		else entries.sort_by_format(cfg_sorttree ? (const char*)cfg_sort_order : p_dbe->get_hierarchy(),0);
		b_sorted = true;
	}
}



ptr_list_t<album_list_window> album_list_window::list_wnd;

class font_notify : public cfg_var_notify
{
	HFONT g_font;

	virtual void on_var_change(const char * name,cfg_var * var)
	{
		if (var==&cfg_font)
		{
			if (g_font!=0)
			{
				unsigned n, count = album_list_window::list_wnd.get_count();
				for (n=0; n<count; n++)
				{
					HWND wnd = album_list_window::list_wnd[n]->wnd_tv;
					if (wnd) uSendMessage(wnd,WM_SETFONT,(WPARAM)0,MAKELPARAM(0,0));
				}
				DeleteObject(g_font);
			}
			
			g_font = uCreateFontIndirect(&(uLOGFONT)cfg_font);
			
			unsigned n, count = album_list_window::list_wnd.get_count();
			for (n=0; n<count; n++)
			{
				HWND wnd = album_list_window::list_wnd[n]->wnd_tv;
				if (wnd) 
				{
					uSendMessage(wnd,WM_SETFONT,(WPARAM)g_font,MAKELPARAM(1,0));
					if (cfg_use_custom_indent)
						TreeView_SetIndent(wnd, cfg_indent);
				}
			}
			
		}
	}
public:
	font_notify(cfg_var * var) : cfg_var_notify(var) {g_font = 0;}
	~font_notify()
	{
		if (g_font) DeleteObject(g_font); g_font=0;
	}
};

static font_notify g_font_notify(&cfg_font);

static void process_entry_recur(metadb_handle * handle,const file_info * info,unsigned level,const ptr_list_t<char> &hierarchy,node * parent, bool format = true)
{
	parent->set_bydir(false);
	parent->add_entry(handle);

	if (level<hierarchy.get_count())
	{		
		const char * field = hierarchy[level];
		if (strchr(field,'%') || strchr(field,'$') || strchr(field,'#'))
		{
			titleformat::g_run(info,g_formatbuf,field,0);
			if (!g_formatbuf.is_empty())
			{
				node * child = parent->add_child(handle,g_formatbuf,g_formatbuf.length());
				if (child)
					process_entry_recur(handle,info,level+1,hierarchy,child);
			}
			else
			{
				process_entry_recur(handle,info,level+1,hierarchy,parent);
			}
		}
		else
		{
			unsigned n;
			for(n=0;;n++)
			{
				const char * val = info->meta_get(field,n);
				if (val==0) break;
				node * child = parent->add_child(handle,val,strlen(val));
				if (child)
					process_entry_recur(handle,info,level+1,hierarchy,child);
			}
		}
	}
}

static bool is_path_separator_custom(char c) {return c=='\\' || c=='|' || c=='/';}

static void process_entry_recur_bydir(metadb_handle * handle,const char * path,node * parent)
{
	parent->set_bydir(true);
	parent->add_entry(handle);

	while(is_path_separator_custom(*path)) path++;

	if (*path)
	{
		unsigned len = 0;
		while(path[len] && !is_path_separator_custom(path[len])) len++;

		node * child = parent->add_child(handle,path,len);
		if (child)
			process_entry_recur_bydir(handle,path+len,child);
	}
}


class enum_callback
{
private:
	ptr_list_t<char> hierarchy;
	bool b_bydir;
public:
	~enum_callback() {deinit();}

	void deinit() {hierarchy.free_all();}

	void run(album_list_window * p_dbe)
	{
		deinit();
		b_bydir = p_dbe->is_bydir();

		if (b_bydir)
		{
			string8_fastalloc path;
			metadb_handle_list list;
			metadb::get()->get_all_entries(list);
			unsigned n;
			for(n=0;n<list.get_count();n++)
			{
				if (list[n]->handle_get_relative_path(path))
				{
					process_entry_recur_bydir(list[n],path,&p_dbe->r00t);
				}
			}
			list.delete_all();
		}
		else
		{
			const char * ptr = p_dbe->get_hierarchy();
			while(*ptr)
			{
				const char * start = ptr;
				while(*ptr && *ptr!='|') ptr++;
				if (ptr>start) hierarchy.add_item(strdup_n(start,ptr-start));
				while(*ptr=='|') ptr++;
			}
			if (hierarchy.get_count()==0)
				hierarchy.add_item(strdup("%_filename%"));

			metadb_handle_list list;
			in_metadb_sync blah;
			metadb::get()->get_all_entries(list);
			unsigned n;
			for(n=0;n<list.get_count();n++)
			{
				const file_info * info = list[n]->handle_query_locked();
				if (info) process_entry_recur(list[n],info,0,hierarchy,&p_dbe->r00t);
			}
			list.delete_all();
		}
	}
};

static enum_callback g_enum_callback;


static void setup_tree(HWND list,HTREEITEM parent,node * ptr,int level,int idx,int max_idx,metadb_handle_list & entries, bool b_sort)
{
	static string8_fastalloc sz_text;	
	sz_text.reset();

	if (cfg_show_numbers2 && max_idx>0)
	{
		int pad = 0;
		while(max_idx>0)
		{
			max_idx/=10;
			pad++;
		}
		char temp1[128],temp2[128];
		sprintf(temp1,"%%0%uu. ",pad);
		sprintf(temp2,temp1,idx+1);
		sz_text += temp2;
	}
		
	sz_text += ptr->get_val();

	if (cfg_show_numbers)
	{
		int num = ptr->get_num_children();
		if (num>0)
		{
			char blah[64];
			sprintf(blah," (%u)",num);
			sz_text += blah;
		}
	}

	HTREEITEM item = TVI_ROOT;


	if (level > 0 || cfg_show_root)
	{
		uTVINSERTSTRUCT is;
		memset(&is,0,sizeof(is));
		is.hParent = parent;
		is.hInsertAfter = TVI_LAST;
		is.item.mask = TVIF_TEXT|TVIF_PARAM|TVIF_STATE;
		is.item.pszText = const_cast<char*>(sz_text.get_ptr());
		is.item.lParam = (int)ptr;
		is.item.state = level < 1 ? TVIS_EXPANDED : 0;
		is.item.stateMask = TVIS_EXPANDED;
		item = uTreeView_InsertItem(list,&is);
	}

	const ptr_list_t<node> & children = ptr->get_children();
	
	unsigned n;

	if (b_sort)
	{

		for(n=0;n<children.get_count();n++)
			entries.add_item(children[n]->get_entries()[0]);

		int * order = (int*)alloca(children.get_count() * sizeof(int));

		entries.sort_by_format_get_order(order,cfg_sort_order,0);

		entries.remove_all();

		for(n=0;n<children.get_count();n++)
			setup_tree(list,item,children[order[n]],level+1,n,children.get_count(),entries,true);

	}
	else
	{
		for(n=0;n<children.get_count();n++)
			setup_tree(list,item,children[n],level+1,n,children.get_count(),entries,false);
	}
}






album_list_window::album_list_window() : wnd_tv(0), r00t(0,0,0,this), initialised(false), view("by artist/album"), treeproc(0), dragging(false), clicked(false), clickpoint(0), p_selection(0), indent_default(0)
{
}

album_list_window::~album_list_window()
{
	if (initialised) 
	{
		list_wnd.remove_item(this);
		initialised = false;
	}
}

void album_list_window::update_all_window_frames()
{
	unsigned n, count = list_wnd.get_count();
	long flags = 0;
	if (cfg_frame == 1) flags |= WS_EX_CLIENTEDGE;
	if (cfg_frame == 2) flags |= WS_EX_STATICEDGE;
	
	for (n=0; n<count; n++)
	{
		HWND wnd = list_wnd[n]->wnd_tv;
		if (wnd)
		{
			uSetWindowLong(wnd, GWL_EXSTYLE, flags);
			SetWindowPos(wnd,0,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER|SWP_FRAMECHANGED);
		}
	}
}

void album_list_window::update_all_colours()
{
			
	unsigned n, count = list_wnd.get_count();
	for (n=0; n<count; n++)
	{
		HWND wnd = list_wnd[n]->wnd_tv;
		if (wnd) 
		{
			list_wnd[n]->update_colours();
		}
	}

}

void album_list_window::update_all_item_heights()
{
			
	unsigned n, count = list_wnd.get_count();
	for (n=0; n<count; n++)
	{
		HWND wnd = list_wnd[n]->wnd_tv;
		if (wnd) 
		{
			list_wnd[n]->update_item_height();
		}
	}

}

void album_list_window::update_all_heights()
{
			
	unsigned n, count = list_wnd.get_count();
	for (n=0; n<count; n++)
	{
		HWND wnd = list_wnd[n]->get_wnd();
		if (wnd) 
		{
			ui_extension_host * host = list_wnd[n]->get_host();
			if (host) 
			{
				host->on_size_limit_change(wnd, ui_extension_flag::SLC_MIN_HEIGHT);
			}
		}
	}
}

void album_list_window::update_all_indents()
{
			
	unsigned n, count = list_wnd.get_count();
	for (n=0; n<count; n++)
	{
		HWND wnd = list_wnd[n]->get_wnd();
		if (wnd) 
		{
			TreeView_SetIndent(list_wnd[n]->wnd_tv, cfg_use_custom_indent ? cfg_indent : list_wnd[n]->indent_default);
		}
	}
}

void album_list_window::update_colours()
{
	TreeView_SetBkColor(wnd_tv, cfg_use_custom_colours ? cfg_colour_back : -1);
	TreeView_SetLineColor(wnd_tv, cfg_use_custom_colours ? cfg_colour_line : -1);
	TreeView_SetTextColor(wnd_tv, cfg_use_custom_colours ? cfg_colour_text : -1);
	
}

void album_list_window::update_item_height()
{
	HFONT font = (HFONT)uSendMessage(wnd_tv, WM_GETFONT, 0, 0);
	int font_height = -1;
	if (cfg_custom_item_height)
	{
		font_height = uGetFontHeight(font) + cfg_item_height; 
		if (font_height < 1) font_height = 1;
	}
	TreeView_SetItemHeight(wnd_tv, font_height);
}

void album_list_window::refresh_tree()
{
	uSendMessage(wnd_tv,WM_SETREDRAW,FALSE,0);

	bool b_sort = !!cfg_sorttree;

	uSendMessage(wnd_tv,TVM_DELETEITEM,0,(long)TVI_ROOT);
	p_selection = 0;
	r00t.reset();
	g_enum_callback.run(this);
	
	//pointers to hierarchy strings kept by nodes for sorting

	if (!b_sort) r00t.sort_children();

	{
		metadb_handle_list entries;//hacks to avoid instantiating shit over and over
		entries.set_mem_logic(mem_block::ALLOC_FAST_DONTGODOWN);

		setup_tree(wnd_tv,TVI_ROOT,&r00t,0,0,0,entries,b_sort);
	}

	uSendMessage(wnd_tv,WM_SETREDRAW,TRUE,0);
}

void album_list_window::build_menu(HMENU menu, unsigned base, POINT pt, bool show_shortcuts, void * & user_data)
{
	if (GetMenuItemCount(menu)) uAppendMenu(menu,(MF_SEPARATOR),0,"");
	
	HMENU menu_view = CreatePopupMenu();
	unsigned n,m = cfg_view_list.get_count();
	string8_fastalloc temp(32);
				
	uAppendMenu(menu_view,MF_STRING|(!stricmp_utf8(directory_structure_view_name,view) ? MF_CHECKED : 0),base+0,directory_structure_view_name);
				
	ptr_list_autodel_t<string_simple> * views  = new ptr_list_autodel_t<string_simple>;
	
	if (views)
	{
		
		views->add_item(new string_simple(directory_structure_view_name));
		
		for(n=0;n<m;n++)
		{
			temp = cfg_view_list.get_name(n);
			string_simple * item = new string_simple(temp.get_ptr());
			
			if (item)
			{
				uAppendMenu(menu_view,MF_STRING|(!stricmp_utf8(temp,view) ? MF_CHECKED : 0),base+views->add_item(item),temp);
			}
			
		}
		uAppendMenu(menu,MF_STRING|MF_POPUP,(UINT)menu_view,"View");
	}
	user_data = views;
	
}

void album_list_window::menu_action(HMENU menu, unsigned base, unsigned id, void * user_data)
{
	ptr_list_autodel_t<string_simple> * views  = (ptr_list_autodel_t<string_simple> *)user_data;

	if (id>=base && views)
	{
		unsigned n = id-base;
		if (n<views->get_count())
		{
			view = views->get_item(n)->get_ptr();
			refresh_tree();
		}
	}
	if (views) delete views;
	
}



static void do_playlist(node * src,bool replace,bool b_new = false)
{
	if (src)
	{
		if (b_new) src->create_new_playlist();
		else src->send_to_playlist(replace);
		if (replace && cfg_autoplay)
		{
			playlist_switcher::get()->reset_playing_playlist();
			play_control::get()->play_start();
		}
	}
}

static void do_autosend_playlist(node * src, string_base & view, bool b_play = false)
{
	if (src)
	{
		string8 playlist_name;
		mem_block_fastalloc<char> extra_items(view.length() + 7);
		extra_items.append("view=",5);
		extra_items.append(view.get_ptr(),view.length());
		extra_items.append("\0",2);
		titleformat::g_run(&file_info_i_full(), playlist_name, cfg_playlist_name, extra_items);
		playlist_name.fix_filename_chars(); //for find playlist
		playlist_switcher::get()->find_or_create_playlist_activate(playlist_name);
		src->sort_entries();
		playlist_oper::get()->add_items(src->get_entries(), true);
		if (b_play && cfg_autoplay)
		{
			playlist_switcher::get()->reset_playing_playlist();
			play_control::get()->play_start();
		}
	}
}

static cfg_dropdown_history history_sort("history_sort",10);

struct edit_view_param
{
	unsigned idx;
	string8 name,value;
	bool b_new;
};

static BOOL CALLBACK EditViewProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		uSetWindowLong(wnd,DWL_USER,lp);
		{
			edit_view_param * ptr = reinterpret_cast<edit_view_param*>(lp);
			uSetDlgItemText(wnd,IDC_NAME,ptr->name);
			uSetDlgItemText(wnd,IDC_VALUE,ptr->value);
		}
		break;
	case WM_COMMAND:
		switch(wp)
		{
		case IDCANCEL:
			EndDialog(wnd,0);
			break;
		case IDOK:
			{
				edit_view_param * ptr = reinterpret_cast<edit_view_param*>(uGetWindowLong(wnd,DWL_USER));
				{
					string8 temp;
					uGetDlgItemText(wnd,IDC_NAME,temp);
					if (temp.is_empty())
					{
						uMessageBox(wnd,"Please enter a valid name.",0,0);
						break;
					}
					unsigned idx_find = cfg_view_list.find_item(temp);
					if (idx_find != -1 && (ptr->b_new || ((idx_find != ptr->idx) && (idx_find != -1))))
					{
						uMessageBox(wnd,"View of this name already exists. Please enter another one.",0,0);
						break;
					}
					ptr->name = temp;
				}
				uGetDlgItemText(wnd,IDC_VALUE,ptr->value);
				EndDialog(wnd,1);

			}
			break;			
		}
		break;
	}
	return FALSE;
}

static bool run_edit_view(edit_view_param & param,HWND parent)
{
	return uDialogBox(IDD_EDIT_VIEW,parent,EditViewProc,reinterpret_cast<LPARAM>(&param)) != 0;
}



static BOOL CALLBACK DialogProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);


LRESULT WINAPI album_list_window::on_message(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{

	switch(msg)
	{
	case WM_CREATE:
		{
			if (!g_metadb) g_metadb = metadb::get();

			list_wnd.add_item(this);

			initialised = true;

			modeless_dialog_manager::add(wnd);

			long flags = 0;
			if (cfg_frame == 1) flags |= WS_EX_CLIENTEDGE;
			else if (cfg_frame == 2) flags |= WS_EX_STATICEDGE;

			wnd_tv = uCreateWindowEx(flags, WC_TREEVIEW, "Album list",
				TVS_HASBUTTONS | TVS_HASLINES | TVS_LINESATROOT  | (cfg_hscroll ? 0 : TVS_NOHSCROLL ) | WS_CHILD | WS_VSCROLL | WS_VISIBLE | WS_TABSTOP, 0, 0, 0, 0,
				wnd, HMENU(IDC_TREE), core_api::get_my_instance(), NULL);
			
			if (wnd_tv)
			{
				indent_default =  TreeView_GetIndent(wnd_tv);

				cfg_font.refresh();

				if (cfg_custom_item_height)
					update_item_height();

				update_colours();

				uSetWindowLong(wnd_tv,GWL_USERDATA,(LPARAM)(this));
				treeproc = (WNDPROC)uSetWindowLong(wnd_tv,GWL_WNDPROC,(LPARAM)(hook_proc));
			}

			if (cfg_populate) refresh_tree();
		}
	break;
	case WM_GETMINMAXINFO:
	{
		LPMINMAXINFO mmi = LPMINMAXINFO(lp);
		mmi->ptMinTrackSize.y = cfg_height;
		return 0;
	}
	case WM_SIZE:
		SetWindowPos(wnd_tv, 0, 0, 0, LOWORD(lp), HIWORD(lp), SWP_NOZORDER);
		break;
/*	case DM_GETDEFID:
		return (DC_HASDEFID<<16|IDOK);
	case WM_GETDLGCODE:
		return DLGC_DEFPUSHBUTTON;*/
//		break;
	case WM_COMMAND:
		switch(wp)
		{
		case IDOK:
			if (GetAsyncKeyState(VK_SHIFT) & KF_UP) do_playlist(p_selection,false);
			else if (GetAsyncKeyState(VK_CONTROL) & KF_UP) do_playlist(p_selection,true,true);
			else do_playlist(p_selection,true);
			return 0;
		case IDC_SEND:
			do_playlist(p_selection,true);
			break;
		case IDC_ADD:
			do_playlist(p_selection,false);
			break;
		case IDC_REMOVE:
			if (p_selection!=0)
			{
				p_selection->remove_from_db();
				uPostMessage(wnd,WM_COMMAND,IDC_REFRESH,0);
			}
			break;
		case IDC_REMOVEDEAD:
			if (p_selection!=0)
			{
				p_selection->remove_dead();
				uPostMessage(wnd,WM_COMMAND,IDC_REFRESH,0);
			}
			break;
		case IDC_REFRESH:
			{
				refresh_tree();
			}
			break;
		}
		break;
		case WM_CONTEXTMENU:
			{
				enum { ID_SEND = 1, ID_ADD, ID_NEW, ID_AUTOSEND, ID_REMOVE, ID_REMOVEDEAD, ID_REFRESH, ID_VIEW_BASE };

				HMENU menu = CreatePopupMenu();

				POINT pt = {GET_X_LPARAM(lp), GET_Y_LPARAM(lp)};
				menu_manager * p_menu_manager = 0;

				unsigned IDM_MANAGER_BASE = 0;
				
				HWND list =wnd_tv;
				
				HTREEITEM treeitem = TreeView_GetSelection(list);
				HTREEITEM treeitem_context = 0;
				
				TVHITTESTINFO ti;
				memset(&ti,0,sizeof(ti));
				
				if (pt.x != -1 && pt.y != -1)
				{
					ti.pt = pt;
					ScreenToClient(list,&ti.pt);
					uSendMessage(list,TVM_HITTEST,0,(long)&ti);
					if (ti.hItem && (ti.flags & TVHT_ONITEM))
					{
						uSendMessage(list,TVM_SELECTITEM,TVGN_CARET,(long)ti.hItem);
						treeitem_context = ti.hItem;
					}
				}
				else
				{
					treeitem_context = TreeView_GetSelection(list);

					RECT rc;
					if (treeitem_context && TreeView_GetItemRect(wnd_tv, treeitem_context, &rc, TRUE))
					{
						MapWindowPoints(wnd_tv, HWND_DESKTOP, (LPPOINT)&rc, 2);

						pt.x = rc.left;
						pt.y = rc.top + (rc.bottom - rc.top) / 2;

					}
					else
					{
						GetMessagePos(&pt);
					}
				}
				
				HMENU menu_view = CreatePopupMenu();
				unsigned n,m = cfg_view_list.get_count();
				string8_fastalloc temp(32);
				
				uAppendMenu(menu_view,MF_STRING|(!stricmp_utf8(directory_structure_view_name,view) ? MF_CHECKED : 0),ID_VIEW_BASE+0,directory_structure_view_name);
				
				ptr_list_autodel_t<string_simple> views;
				
				views.add_item(new string_simple(directory_structure_view_name));
				
				for(n=0;n<m;n++)
				{
					temp = cfg_view_list.get_name(n);
					string_simple * item = new string_simple(temp.get_ptr());
					
					if (item)
					{
						uAppendMenu(menu_view,MF_STRING|(!stricmp_utf8(temp,view) ? MF_CHECKED : 0),ID_VIEW_BASE+views.add_item(item),temp);
					}
					
				}
				
				
				IDM_MANAGER_BASE = ID_VIEW_BASE + views.get_count();
				
				uAppendMenu(menu,MF_STRING|MF_POPUP,(UINT)menu_view,"View");
				
				uAppendMenu(menu,MF_STRING,ID_REFRESH,"Refresh tree");

				bool show_shortcuts = get_host()->get_show_shortcuts();
				
				if (p_selection && treeitem_context)
				{
					uAppendMenu(menu,MF_SEPARATOR,0,"");
					uAppendMenu(menu,MF_STRING,ID_SEND,(show_shortcuts ? "&Send to playlist\tEnter" : "&Send to playlist"));
					uAppendMenu(menu,MF_STRING,ID_ADD,show_shortcuts ? "&Add to playlist\tShift+Enter" : "&Add to playlist");
					uAppendMenu(menu,MF_STRING,ID_NEW,show_shortcuts ?"Send to &new playlist\tCtrl+Enter" : "Send to &new playlist");
					uAppendMenu(menu,MF_STRING,ID_AUTOSEND,"Send to &autosend playlist");
					uAppendMenu(menu,MF_STRING,ID_REMOVE,"&Remove from database");
					uAppendMenu(menu,MF_STRING,ID_REMOVEDEAD,"Remove &dead entries (slow)");
					uAppendMenu(menu,MF_SEPARATOR,0,"");
					
					p_menu_manager = menu_manager::create();
					p_selection->sort_entries();
					
					if (p_menu_manager)
					{
						p_menu_manager->init_context(p_selection->get_entries(), 0);
						
						p_menu_manager->v_win32_build_menu(menu,IDM_MANAGER_BASE,-1);
						p_menu_manager->v_win32_auto_mnemonics(menu);
					}
				}
				
				int cmd = TrackPopupMenu(menu,TPM_RIGHTBUTTON|TPM_NONOTIFY|TPM_RETURNCMD,pt.x,pt.y,0,get_wnd(),0);
				DestroyMenu(menu);
				
				if (cmd)
				{
					if (p_menu_manager && (unsigned)cmd>=IDM_MANAGER_BASE)
					{
						p_menu_manager->execute_by_id(cmd - IDM_MANAGER_BASE);
					}
					else if (cmd>=ID_VIEW_BASE)
					{
						unsigned n = cmd-ID_VIEW_BASE;
						if (n<views.get_count())
						{
							view = views[n]->get_ptr();
							refresh_tree();
						}
					}
					else if (cmd<ID_VIEW_BASE)
					{
						unsigned cmd2 = 0;
						switch(cmd)
						{
						case ID_NEW:
							do_playlist(p_selection,true,true);
							break;
						case ID_SEND:
							do_playlist(p_selection,true);
							break;
						case ID_ADD:
							do_playlist(p_selection,false);
							break;
						case ID_AUTOSEND:
							do_autosend_playlist(p_selection, view, true);
							break;
						case ID_REMOVE:
							cmd2 = IDC_REMOVE;
							break;
						case ID_REMOVEDEAD:
							cmd2 = IDC_REMOVEDEAD;
							break;
						case ID_REFRESH:
							cmd2 = IDC_REFRESH;
							break;
						}
						if (cmd2) uSendMessage(get_wnd(),WM_COMMAND,cmd2,0);
					}
				}
				
				if (p_menu_manager)
				{
					p_menu_manager->service_release();
					p_menu_manager = 0;
				}
				
	/*			if (treeitem_context && (treeitem_context != treeitem) && cfg_autosend)
					TreeView_SelectItem(wnd_tv,treeitem);*/


			}
			return 0;
	case WM_NOTIFY:
		{
			LPNMHDR hdr = (LPNMHDR)lp;

			switch (hdr->idFrom) 
			{

			case IDC_TREE:
				{
				if (hdr->code == TVN_SELCHANGEDA)
				{
					
					LPNMTREEVIEWA param = (LPNMTREEVIEWA)hdr;
					p_selection = (node*)param->itemNew.lParam;
					if(p_selection)
						p_selection->sort_entries();
						if(p_selection->get_num_entries()>0 && !albumart_focus(p_selection->get_entries().get_item(0)))
							console::info("Could not find or message foo_uie_albumart.");

					if (cfg_autosend &&( param->action == TVC_BYMOUSE || param->action == TVC_BYKEYBOARD) )
					{
						do_autosend_playlist(p_selection, view);
					}
					
				}
				else if (hdr->code == TVN_SELCHANGEDW)
				{
					LPNMTREEVIEWW param = (LPNMTREEVIEWW)hdr;
					p_selection = (node*)param->itemNew.lParam;
					if(p_selection)
						p_selection->sort_entries();
						if(p_selection->get_num_entries()>0 && !albumart_focus(p_selection->get_entries().get_item(0)))
							console::info("Could not find or message foo_uie_albumart.");

					if (cfg_autosend && (param->action == TVC_BYMOUSE || param->action == TVC_BYKEYBOARD) )
					{
						do_autosend_playlist(p_selection, view);
					}
				}
				else if(hdr->code == NM_SETFOCUS)
				{
					if(p_selection && p_selection->get_num_entries()>0 && !albumart_focus(p_selection->get_entries().get_item(0)))
						console::info("Could not find or message foo_uie_albumart.");
				}
				else if(hdr->code == NM_KILLFOCUS)
				{
					if(!albumart_focus(NULL))
						console::info("Could not find or message foo_uie_albumart.");
				}
			}
			break;
			}
		
		} break;
	case WM_DESTROY:
		modeless_dialog_manager::remove(wnd);
		wnd_tv=0;
		r00t.reset();
		if (initialised) 
		{
			list_wnd.remove_item(this);
			initialised = false;
		}
		break;
	}
	return uDefWindowProc(wnd, msg, wp, lp);
}

LRESULT WINAPI album_list_window::hook_proc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	album_list_window * p_this;
	LRESULT rv;
	
	p_this = reinterpret_cast<album_list_window*>(uGetWindowLong(wnd,GWL_USERDATA));
	
	rv = p_this ? p_this->on_hook(wnd,msg,wp,lp) : uDefWindowProc(wnd, msg, wp, lp);

	return rv;
}

static bool test_point_distance(DWORD pt1,DWORD pt2,int test)
{
	int dx = (short)LOWORD(pt1) - (short)LOWORD(pt2),
		dy = (short)HIWORD(pt1) - (short)HIWORD(pt2);
	return dx*dx + dy*dy > test*test;	
}

LRESULT WINAPI album_list_window::on_hook(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	static process_char = true;

	switch(msg)
	{
	case WM_KEYDOWN:
		if (get_host()->get_keyboard_shortcuts_enabled() && wp != VK_LEFT && wp != VK_RIGHT && cfg_keyb && keyboard_shortcut_manager::get()->on_keydown_auto(wp)) {process_char = false; break;}
		else if (wp == VK_TAB)
		{
			ui_extension::g_on_tab(wnd);
		}
		process_char = true;
		break;
	case WM_SYSKEYDOWN:
		if (get_host()->get_keyboard_shortcuts_enabled() && wp != VK_LEFT && wp != VK_RIGHT && cfg_keyb && keyboard_shortcut_manager::get()->on_keydown_auto(wp)) {process_char = false; break;}
		process_char = true;
		break;
	case  WM_CHAR:
		if (cfg_keyb && !process_char)
		{
			process_char = true;
			return 0;
		}
		break;
	case WM_GETDLGCODE:
		{
			MSG * msg = (LPMSG)lp;
			if (msg && cfg_keyb)
			{
				// let dialog manager handle it, otherwise to kill ping we have to process WM_CHAR to return 0 on wp == 0xd and 0xa
				if (!((msg->message == WM_KEYDOWN || msg->message == WM_KEYUP) && msg->wParam == VK_RETURN)) return DLGC_WANTMESSAGE;
			}

/*			if (cfg_keyb)
				;*///return DLGC_DEFPUSHBUTTON;
		}
		break;
	case WM_LBUTTONUP:
		clicked = false;
		break;
	case WM_LBUTTONDOWN:
		clicked = true;
		clickpoint = lp;
		break;
	case WM_MOUSEMOVE:
		{
			if (!(wp&MK_LBUTTON)/* || (wp&MK_RBUTTON)*/) {dragging = false;clicked = false;}
			else if (!dragging && clicked && test_point_distance(clickpoint,lp,5))
			{
				TVHITTESTINFO ti;
				memset(&ti,0,sizeof(ti));
				ti.pt.x = (short)LOWORD(clickpoint);
				ti.pt.y = (short)HIWORD(clickpoint);
				uSendMessage(wnd,TVM_HITTEST,0,(LPARAM)&ti);
				
				if (ti.flags & TVHT_ONITEM)
				{
//					HTREEITEM item = TreeView_GetSelection(wnd);
					if (ti.hItem) uSendMessage(wnd,TVM_SELECTITEM,TVGN_CARET,(LPARAM)ti.hItem);
					
					if (p_selection)
					{
						p_selection->sort_entries();
						dragging = true;
						IDataObject * pDataObject = playlist_oper::get()->create_dataobject(p_selection->get_entries());
						if (pDataObject)
						{
							DWORD blah;
							IDropSource_albumlist * p_IDropSource_albumlist = new IDropSource_albumlist(wnd_tv);
							DoDragDrop(pDataObject,p_IDropSource_albumlist,DROPEFFECT_COPY,&blah);
							delete p_IDropSource_albumlist;
							pDataObject->Release();
						}
						dragging = false;
						clicked = false;
					}
					
//					uSendMessage(wnd,TVM_SELECTITEM,TVGN_CARET,(long)item);
				}
			}
		}
		break;
	case WM_MBUTTONDOWN:
		{
			TVHITTESTINFO ti;
			memset(&ti,0,sizeof(ti));
			
			ti.pt.x = GET_X_LPARAM(lp);
			ti.pt.y = GET_Y_LPARAM(lp);
			uSendMessage(wnd,TVM_HITTEST,0,(long)&ti);
			if (ti.flags & TVHT_ONITEM)
			{
				int action = cfg_middle;
				if (action)
				{
					POINT pt = {(short)LOWORD(lp),(short)HIWORD(lp)};
					TVHITTESTINFO ti;
					memset(&ti,0,sizeof(ti));
					ti.pt = pt;
					uSendMessage(wnd,TVM_HITTEST,0,(LPARAM)&ti);
					if (ti.hItem)
					{
						uSendMessage(wnd,TVM_SELECTITEM,TVGN_CARET,(LPARAM)ti.hItem);
						switch(action)
						{
						case 1:
							do_playlist(p_selection,true);
							break;
						case 2:
							do_playlist(p_selection,false);
							break;
						case 3:
							do_playlist(p_selection,true,true);
							break;
						case 4:
							do_autosend_playlist(p_selection, view, true);
							break;
						}
						
					}
				}
			}
		}
		
		break;
	case WM_LBUTTONDBLCLK:
		if (p_selection && p_selection->get_num_entries()>0)
		{
			TVHITTESTINFO ti;
			memset(&ti,0,sizeof(ti));
			
			ti.pt.x = GET_X_LPARAM(lp);
			ti.pt.y = GET_Y_LPARAM(lp);
			uSendMessage(wnd,TVM_HITTEST,0,(long)&ti);
			if (ti.flags & TVHT_ONITEM)
			{
				
				switch(cfg_dblclk)
				{
				case 0:
					if (p_selection==0 || p_selection->get_num_children()>0) break;
					do_playlist(p_selection,true);
					return 0;
				case 1:
					do_playlist(p_selection,true);
					return 0;
				case 2:
					do_playlist(p_selection,false);
					return 0;
				case 3:
					do_playlist(p_selection,true,true);
					return 0;
				case 4:
					do_autosend_playlist(p_selection, view, true);
					return 0;
				}
				break;
			}	
		}
	}
	return uCallWindowProc(treeproc,wnd,msg,wp,lp);
}

void album_list_window::get_config(cfg_var::write_config_callback * out)
{
	out->write_string(view);
}




bool colour_picker(HWND wnd, cfg_int & out, COLORREF custom = 0)
{
	bool rv = false;
	COLORREF COLOR = out;
	COLORREF COLORS[16] = {custom,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	if (uChooseColor(&COLOR, wnd, &COLORS[0]))
	{
		out = COLOR;
		rv = true;
	}
	return rv;
}

cfg_int cfg_child("prefs_child",0);


class preferences_tab
{
public:
	virtual HWND create(HWND wnd)=0;
	virtual const char * get_name()=0;
};


class tab_general : public preferences_tab
{
	static bool initialised;
	static BOOL CALLBACK ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
	{
		switch(msg)
		{

		case WM_INITDIALOG:

			{
				HWND list = uGetDlgItem(wnd,IDC_VIEWS);
				unsigned n,m = cfg_view_list.get_count();
				string8_fastalloc temp;
				for(n=0;n<m;n++)
				{
					cfg_view_list.format_display(n,temp);
					uSendMessageText(list,LB_ADDSTRING,0,temp);
				}
			}


			history_sort.add_item(cfg_sort_order);
			history_sort.setup_dropdown(GetDlgItem(wnd,IDC_SORT_SPEC));

			uSetDlgItemText(wnd,IDC_SORT_ORDER,cfg_sort_order);
			uSendDlgItemMessage(wnd,IDC_SHOW_NUMBERS,BM_SETCHECK,cfg_show_numbers,0);
			uSendDlgItemMessage(wnd,IDC_SHOW_NUMBERS2,BM_SETCHECK,cfg_show_numbers2,0);
			
			{
				HWND list = uGetDlgItem(wnd,IDC_DBLCLK);
				uSendMessageText(list,CB_ADDSTRING,0,"Expand/collapse (default)");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Add to playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to new playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to autosend playlist");
				uSendMessage(list,CB_SETCURSEL,cfg_dblclk,0);
				
				list = uGetDlgItem(wnd,IDC_MIDDLE);
				uSendMessageText(list,CB_ADDSTRING,0,"None");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Add to playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to new playlist");
				uSendMessageText(list,CB_ADDSTRING,0,"Send to autosend playlist");
				uSendMessage(list,CB_SETCURSEL,cfg_middle,0);
			}


//			uSendDlgItemMessage(wnd,IDC_AUTOPLAY,BM_SETCHECK,cfg_autoplay,0);			
			uSendDlgItemMessage(wnd,IDC_USE_CUSTOMSORT,BM_SETCHECK,cfg_sorttree,0);

			EnableWindow(uGetDlgItem(wnd,IDC_SORT_ORDER),cfg_sorttree);

//			uSetDlgItemText(wnd,IDC_RESULTS_SORT,cfg_results_sort_spec);
//			uSetDlgItemText(wnd,IDC_HIERARCHYNEW,cfg_hierarchy);
			uSetDlgItemText(wnd,IDC_PLAYLIST_NAME,cfg_playlist_name);

			uSendDlgItemMessage(wnd,IDC_AUTO_SEND,BM_SETCHECK,cfg_autosend,0);

			SetDlgItemInt(wnd,IDC_HEIGHT,cfg_height, FALSE);
			uSendDlgItemMessage(wnd,IDC_SPIN,UDM_SETRANGE32,0,999);
			initialised=true;

			break;
		case WM_COMMAND:
			switch(wp)
			{
			case IDC_USE_CUSTOMSORT:
				cfg_sorttree = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				EnableWindow(uGetDlgItem(wnd,IDC_SORT_ORDER),cfg_sorttree);
				break;
			case IDC_MIDDLE | (CBN_SELCHANGE<<16):
				cfg_middle = uSendMessage((HWND)lp,CB_GETCURSEL,0,0);
				break;
			case IDC_DBLCLK | (CBN_SELCHANGE<<16):
				cfg_dblclk = uSendMessage((HWND)lp,CB_GETCURSEL,0,0);
				break;
			case IDC_AUTO_SEND:
				cfg_autosend = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case (CBN_KILLFOCUS<<16)|IDC_SORT_ORDER:
				cfg_sort_order = string_utf8_from_window((HWND)lp);
				break;
//			case IDC_TAGZHELP:
//				standard_commands::main_titleformat_help();
//				break;
			case (EN_KILLFOCUS<<16)|IDC_PLAYLIST_NAME:
				cfg_playlist_name = string_utf8_from_window((HWND)lp);
				break;
			case IDC_VIEWS | (LBN_DBLCLK<<16):
				{
					HWND list = (HWND)lp;
					unsigned idx = uSendMessage(list,LB_GETCURSEL,0,0);
					if (idx != LB_ERR)
					{
						edit_view_param p;
						p.b_new = false;
						p.idx = idx;
						p.name = cfg_view_list.get_name(idx);
						p.value = cfg_view_list.get_value(idx);
						if (run_edit_view(p,wnd))
						{
							string8 temp;
							cfg_view_list.modify_item(idx,p.name,p.value);
							cfg_view_list.format_display(idx,temp);
							uSendMessage(list,LB_DELETESTRING,idx,0);
							uSendMessageText(list,LB_INSERTSTRING,idx,temp);
							uSendMessageText(list,LB_SETCURSEL,idx,0);
						}
					}
				}
				break;
			case IDC_VIEW_UP:
				{
					HWND list = uGetDlgItem(wnd,IDC_VIEWS);
					unsigned idx = uSendMessage(list,LB_GETCURSEL,0,0);
					if (idx != LB_ERR && idx>0)
					{
						uSendMessage(list,LB_DELETESTRING,idx,0);
						cfg_view_list.swap(idx,idx-1);
						string8 temp;
						cfg_view_list.format_display(idx-1,temp);
						uSendMessageText(list,LB_INSERTSTRING,idx-1,temp);
						uSendMessage(list,LB_SETCURSEL,idx-1,0);
					}
				}
				break;
			case IDC_VIEW_DOWN:
				{
					HWND list = uGetDlgItem(wnd,IDC_VIEWS);
					unsigned idx = uSendMessage(list,LB_GETCURSEL,0,0);
					if (idx != LB_ERR && idx+1<cfg_view_list.get_count())
					{
						uSendMessage(list,LB_DELETESTRING,idx,0);
						cfg_view_list.swap(idx,idx+1);
						string8 temp;
						cfg_view_list.format_display(idx+1,temp);
						uSendMessageText(list,LB_INSERTSTRING,idx+1,temp);
						uSendMessage(list,LB_SETCURSEL,idx+1,0);
					}
				}
				break;
			case IDC_VIEW_DELETE:
				{
					HWND list = uGetDlgItem(wnd,IDC_VIEWS);
					unsigned idx = uSendMessage(list,LB_GETCURSEL,0,0);
					if (idx != LB_ERR)
					{
						cfg_view_list.remove_item(idx);
						uSendDlgItemMessage(wnd,IDC_VIEWS,LB_DELETESTRING,idx,0);
					}
				}
				break;
			case IDC_VIEW_NEW:
				{
					edit_view_param p;
					p.b_new = true;
					p.idx = -1;
					if (run_edit_view(p,wnd))
					{
						HWND list = uGetDlgItem(wnd,IDC_VIEWS);
						unsigned n = cfg_view_list.add_item(p.name,p.value);
						string8 temp;
						cfg_view_list.format_display(n,temp);
						uSendMessageText(list,LB_ADDSTRING,0,temp);
						uSendMessage(list,LB_SETCURSEL,n,0);
					}
				}
				break;
			case IDC_VIEW_RESET:
				{
					cfg_view_list.reset();
					HWND list = uGetDlgItem(wnd,IDC_VIEWS);
					uSendMessage(list,LB_RESETCONTENT,0,0);
					unsigned n,m = cfg_view_list.get_count();
					string8_fastalloc temp;
					for(n=0;n<m;n++)
					{
						cfg_view_list.format_display(n,temp);
						uSendMessageText(list,LB_ADDSTRING,0,temp);
					}
				}
				break;
			case IDC_FONT:
				{
					uLOGFONT temp = cfg_font;
					if (uChooseFont(&temp,wnd))
						cfg_font = temp;
				}
				break;
			}
			break;
		case WM_DESTROY:
			history_sort.add_item(cfg_sort_order);
			initialised=false;
			break;
		}
		return 0;
	}
public:
	virtual HWND create(HWND wnd) {return uCreateDialog(IDD_CONFIG,wnd,ConfigProc);}
	virtual const char * get_name() {return "General";}

} g_config_general;

class tab_advanced : public preferences_tab
{
	static bool initialised;
	static BOOL CALLBACK ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
	{
		switch(msg)
		{

		case WM_INITDIALOG:

			uSendDlgItemMessage(wnd,IDC_SHOW_NUMBERS,BM_SETCHECK,cfg_show_numbers,0);
			uSendDlgItemMessage(wnd,IDC_SHOW_NUMBERS2,BM_SETCHECK,cfg_show_numbers2,0);
			
			{
				HWND list = uGetDlgItem(wnd,IDC_FRAME);

				uSendMessageText(list,CB_ADDSTRING,0,"None");
				uSendMessageText(list,CB_ADDSTRING,0,"Sunken");
				uSendMessageText(list,CB_ADDSTRING,0,"Grey");

				uSendMessage(list,CB_SETCURSEL,cfg_frame,0);
			}



			uSendDlgItemMessage(wnd,IDC_USE_CUSTOM_COLOURS,BM_SETCHECK,cfg_use_custom_colours,0);
			uSendDlgItemMessage(wnd,IDC_KEYB,BM_SETCHECK,cfg_keyb,0);
			uSendDlgItemMessage(wnd,IDC_POPULATE,BM_SETCHECK,cfg_populate,0);
			uSendDlgItemMessage(wnd,IDC_HSCROLL,BM_SETCHECK,cfg_hscroll,0);
			uSendDlgItemMessage(wnd,IDC_SHOW_ROOT,BM_SETCHECK,cfg_show_root,0);
			uSendDlgItemMessage(wnd,IDC_AUTOPLAY,BM_SETCHECK,cfg_autoplay,0);

			uSendDlgItemMessage(wnd,IDC_USE_INDENT,BM_SETCHECK,cfg_use_custom_indent,0);
			
			{
				HWND wnd_indent = GetDlgItem(wnd, IDC_INDENT);
				
				EnableWindow(wnd_indent, cfg_use_custom_indent);
				EnableWindow(GetDlgItem(wnd, IDC_INDENT_SPIN), cfg_use_custom_indent);
				if (cfg_use_custom_indent)
					SetDlgItemInt(wnd,IDC_INDENT,cfg_indent, TRUE);
				else
					uSendMessageText(wnd_indent, WM_SETTEXT, 0, "");
				
				uSendDlgItemMessage(wnd,IDC_INDENT_SPIN,UDM_SETRANGE32,0,999);
			}
			
			{
				uSendDlgItemMessage(wnd,IDC_USE_ITEM_HEIGHT,BM_SETCHECK,cfg_custom_item_height,0);

				HWND wnd_indent = GetDlgItem(wnd, IDC_ITEM_HEIGHT);
				
				EnableWindow(wnd_indent, cfg_custom_item_height);
				EnableWindow(GetDlgItem(wnd, IDC_ITEM_HEIGHT_SPIN), cfg_custom_item_height);

				if (cfg_custom_item_height)
					SetDlgItemInt(wnd,IDC_ITEM_HEIGHT,cfg_item_height, TRUE);
				else
					uSendMessageText(wnd_indent, WM_SETTEXT, 0, "");
				
				uSendDlgItemMessage(wnd,IDC_ITEM_HEIGHT_SPIN,UDM_SETRANGE32,-99,99);
			}


			SetDlgItemInt(wnd,IDC_HEIGHT,cfg_height, FALSE);
			uSendDlgItemMessage(wnd,IDC_SPIN,UDM_SETRANGE32,0,999);
			initialised=true;

			break;
		case WM_COMMAND:
			switch(wp)
			{
			case IDC_USE_CUSTOM_COLOURS:
				cfg_use_custom_colours = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				album_list_window::update_all_colours();
				break;
			case IDC_POPULATE:
				cfg_populate = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_BACK_COLOUR:
				colour_picker(wnd, cfg_colour_back, GetSysColor(COLOR_WINDOW));
				album_list_window::update_all_colours();
				break;
			case IDC_TEXT_COLOUR:
				colour_picker(wnd, cfg_colour_text, GetSysColor(COLOR_WINDOWTEXT));
				album_list_window::update_all_colours();
				break;
			case IDC_LINE_COLOUR:
				colour_picker(wnd, cfg_colour_line, GetSysColor(COLOR_WINDOWFRAME));
				album_list_window::update_all_colours();
				break;
			case IDC_USE_CUSTOMSORT:
				cfg_sorttree = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				EnableWindow(uGetDlgItem(wnd,IDC_SORT_ORDER),cfg_sorttree);
				break;
			case IDC_SHOW_NUMBERS:
				cfg_show_numbers = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_AUTO_SEND:
				cfg_autosend = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_HSCROLL:
				cfg_hscroll = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_SHOW_ROOT:
				cfg_show_root = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_AUTOPLAY:
				cfg_autoplay = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_KEYB:
				cfg_keyb = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case IDC_SHOW_NUMBERS2:
				cfg_show_numbers2 = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
				break;
			case (CBN_SELCHANGE<<16)|IDC_FRAME:
				{
					cfg_frame = uSendMessage((HWND)lp,CB_GETCURSEL,0,0);
					album_list_window::update_all_window_frames();
				}
				break;
			case (EN_CHANGE<<16)|IDC_HEIGHT:
				{
					if (initialised)
					{
						BOOL result;
						unsigned new_height = GetDlgItemInt(wnd, IDC_HEIGHT, &result, FALSE);
						if (result) 
						{
							if (new_height > 999) new_height = 999;
							cfg_height = new_height;
							album_list_window::update_all_heights();
						}
					}
				}
				break;
			case (EN_CHANGE<<16)|IDC_ITEM_HEIGHT:
				{
					if (initialised && cfg_custom_item_height)
					{
						BOOL result;
						int new_height = GetDlgItemInt(wnd, IDC_ITEM_HEIGHT, &result, TRUE);
						if (result) 
						{
							if (new_height > 99) new_height = 99;
							if (new_height < -99) new_height = -99;
							cfg_item_height = new_height;
							album_list_window::update_all_item_heights();
						}
					}
				}
				break;
			case IDC_USE_ITEM_HEIGHT:
				{
					cfg_custom_item_height = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
					HWND wnd_indent = GetDlgItem(wnd, IDC_ITEM_HEIGHT);
					
					EnableWindow(wnd_indent, cfg_custom_item_height);
					EnableWindow(GetDlgItem(wnd, IDC_ITEM_HEIGHT_SPIN), cfg_custom_item_height);

					if (cfg_custom_item_height)
						SetDlgItemInt(wnd,IDC_ITEM_HEIGHT,cfg_item_height, TRUE);
					else
						uSendMessageText(wnd_indent, WM_SETTEXT, 0, "");
				
					album_list_window::update_all_item_heights();
				}
				break;
			case IDC_USE_INDENT:
				{
					cfg_use_custom_indent = uSendMessage((HWND)lp,BM_GETCHECK,0,0);
					HWND wnd_indent = GetDlgItem(wnd, IDC_INDENT);
					
					EnableWindow(wnd_indent, cfg_use_custom_indent);
					EnableWindow(GetDlgItem(wnd, IDC_INDENT_SPIN), cfg_use_custom_indent);
					if (cfg_use_custom_indent)
						SetDlgItemInt(wnd,IDC_INDENT,cfg_indent, TRUE);
					else
						uSendMessageText(wnd_indent, WM_SETTEXT, 0, "");

					album_list_window::update_all_indents();
				}
				break;
			case (EN_CHANGE<<16)|IDC_INDENT:
				{
					if (initialised && cfg_use_custom_indent)
					{
						BOOL result;
						unsigned new_height = GetDlgItemInt(wnd, IDC_INDENT, &result, TRUE);
						if (result) 
						{
							if (new_height > 999) new_height = 999;
							cfg_indent = new_height;
							album_list_window::update_all_indents();
						}
					}
				}
				break;
			case IDC_FONT:
				{
					uLOGFONT temp = cfg_font;
					if (uChooseFont(&temp,wnd))
						cfg_font = temp;
				}
				break;
			}
			break;
		case WM_DESTROY:
			initialised=false;
			break;
		}
		return 0;
	}
public:
	virtual HWND create(HWND wnd) {return uCreateDialog(IDD_ADVANCED,wnd,ConfigProc);}
	virtual const char * get_name() {return "Advanced";}

} g_config_advanced;

static preferences_tab * g_tabs[] = 
{
	&g_config_general,
	&g_config_advanced,
};


class config_albumlist : public config
{
	
	static HWND child;

	static void make_child(HWND wnd)
	{
		if (child)
		{
			ShowWindow(child, SW_HIDE);
			DestroyWindow(child);
			child=0;
		}

		HWND wnd_tab = GetDlgItem(wnd, IDC_TAB1);
		
		RECT tab;
		
		GetWindowRect(wnd_tab,&tab);
		MapWindowPoints(HWND_DESKTOP, wnd, (LPPOINT)&tab, 2);
		
		TabCtrl_AdjustRect(wnd_tab,FALSE,&tab);
		
		unsigned count = tabsize(g_tabs);
		if ((unsigned)cfg_child >= count) cfg_child = 0;

		if ((unsigned)cfg_child < count && cfg_child >= 0)
		{
			child = g_tabs[cfg_child]->create(wnd);
		}
		
		SetWindowPos(child,HWND_TOP,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
		
		if (child) 
		{
			uxtheme_handle * p_uxtheme = uxtheme_handle::create();//dont use anywhere else, so no point is creating a global handle
			if (p_uxtheme)
			{
				p_uxtheme->EnableThemeDialogTexture(child, ETDT_ENABLETAB);
				p_uxtheme->release();
			}
		}

		SetWindowPos(child, 0, tab.left, tab.top, tab.right-tab.left, tab.bottom-tab.top, SWP_NOZORDER);
		SetWindowPos(wnd_tab,HWND_BOTTOM,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
		
		ShowWindow(child, SW_NORMAL);
		UpdateWindow(child);
	}
	
	static BOOL CALLBACK ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				uTCITEM tabs;
				memset(&tabs, 0, sizeof(tabs));
				
				HWND wnd_tab = GetDlgItem(wnd, IDC_TAB1);
				
				tabs.mask = TCIF_TEXT;

				unsigned n, count = tabsize(g_tabs);


				for (n=0; n<count; n++)
				{
					uTabCtrl_InsertItemText(wnd_tab, n, g_tabs[n]->get_name());
				}
				
				TabCtrl_SetCurSel(wnd_tab, cfg_child);
				
				
				make_child(wnd);
				
			}
			break;
			
			
		case WM_NOTIFY:
			switch (((LPNMHDR)lp)->idFrom)
			{
			case IDC_TAB1:
				switch (((LPNMHDR)lp)->code)
				{
				case TCN_SELCHANGE:
					{
						cfg_child = TabCtrl_GetCurSel(GetDlgItem(wnd, IDC_TAB1));
						make_child(wnd);
					}
					break;
				}
				break;
			}
			break;
		case WM_PARENTNOTIFY:
			switch(wp)
			{
			case WM_DESTROY:
				{
					if (child && (HWND)lp == child) child = 0;
				}
				break;	
			}
			break;
		}
		return 0;
	}
public:
	HWND create(HWND parent)
	{
		return uCreateDialog(IDD_HOST,parent,ConfigProc);
	}
	const char * get_name() {return "Album list panel";}
	const char * get_parent_name() {return "Components";}
};

HWND config_albumlist::child = 0;


static service_factory_single_t<config,config_albumlist> foo3;

bool tab_general::initialised = false;
bool tab_advanced::initialised = false;

/*
const GUID & album_list_window::get_extension_guid() const
{
	return extension_guid;
}*/

void album_list_window::get_name(string_base & out)const
{
	out.set_string("Album list");
}
void album_list_window::get_category(string_base & out)const
{
	out.set_string("Panels");
}

void album_list_window::set_config(cfg_var::read_config_helper * config)
{
		config->read_string(view);
}


// {606E9CDD-45EE-4c3b-9FD5-49381CEBE8AE}
const GUID album_list_window::extension_guid = 
{ 0x606e9cdd, 0x45ee, 0x4c3b, { 0x9f, 0xd5, 0x49, 0x38, 0x1c, 0xeb, 0xe8, 0xae } };


ui_extension_factory<album_list_window> blah;
