#include "uie_albumart.h"

DECLARE_COMPONENT_VERSION("Album Art Panel","0.045","A Columns_UI extension by Nathan Kallus.")

HWND uie_albumart::create_window(HWND wnd_parent)
{
	static const char g_class_name[] = "foo_uie_albumart.dll window class";
	uWNDCLASS wc;
	memset(&wc,0,sizeof(wc));
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
	wc.lpfnWndProc = host_proc;
	wc.hInstance = core_api::get_my_instance();
	wc.hCursor = LoadCursor(0,IDC_ARROW);
	wc.hbrBackground = 0;
	wc.lpszClassName = g_class_name;
	uRegisterClass(&wc);
	return uCreateWindowEx(WS_EX_STATICEDGE,
		g_class_name,"foo_uie_albumart",
		WS_CHILD,
		CW_USEDEFAULT,CW_USEDEFAULT,
		0,0,
		wnd_parent,0,core_api::get_my_instance(),this);
}

LRESULT WINAPI uie_albumart::host_proc(HWND wnd1,UINT msg,WPARAM wp,LPARAM lp){
	uie_albumart * p_this;
	
	if(msg == WM_NCCREATE)
	{
		p_this = (uie_albumart *)((CREATESTRUCT *)(lp))->lpCreateParams; //retrieve pointer to class
		uSetWindowLong(wnd1, GWL_USERDATA, (LPARAM)p_this);//store it for future use
	}else{
		p_this = reinterpret_cast<uie_albumart*>(uGetWindowLong(wnd1,GWL_USERDATA));//if isnt wm_create, retrieve pointer to class
	}
	return p_this ? p_this->on_message(wnd1, msg, wp, lp) : uDefWindowProc(wnd1, msg, wp, lp);
}

LRESULT WINAPI uie_albumart::on_message(HWND wnd1,UINT msg,WPARAM wp,LPARAM lp){
	switch(msg)
	{
	case WM_CREATE:
		GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

		foo_uie_albumart_playlist_notify.add_window(wnd1);
		foo_uie_albumart_playback_notify.add_window(wnd1);

		foo_uie_albumart_src_list_notify.add_window(wnd1);
		foo_uie_albumart_centeralbum_notify.add_window(wnd1);
		foo_uie_albumart_expandalbum_notify.add_window(wnd1);
		foo_uie_albumart_shrinkalbum_notify.add_window(wnd1);
		foo_uie_albumart_selected_notify.add_window(wnd1);
		foo_uie_albumart_usebg_notify.add_window(wnd1);
		foo_uie_albumart_bgcol_notify.add_window(wnd1);
		foo_uie_albumart_aspectratio_notify.add_window(wnd1);
		foo_uie_albumart_pixelborder_notify.add_window(wnd1);
		foo_uie_albumart_interpolationmode_notify.add_window(wnd1);
		foo_uie_albumart_usecontrol_notify.add_window(wnd1);
		foo_uie_albumart_minheight_notify.add_window(wnd1);
		foo_uie_albumart_resetsource_notify.add_window(wnd1);
		foo_uie_albumart_padding_notify.add_window(wnd1);

		image_locked=false;
		crnsource=0;
		if(!plytrack){
			metadb_handle * track=playlist_oper::get()->get_item(playlist_oper::get()->get_now_playing());
			if(track){
				set_handle(plytrack,track);
				track->handle_release();
			}
		}
		if(!seltrack){
			metadb_handle * track=playlist_oper::get()->get_focus_item();
			if(track){
				set_handle(seltrack,track);
				track->handle_release();
			}
		}
		if(img.is_empty() && set_image())
			redraw();
		break;
	case WM_DESTROY:
		GdiplusShutdown(gdiplusToken);
		if(seltrack) seltrack->handle_release();
		if(plytrack) plytrack->handle_release();
		if(crntrack) crntrack->handle_release();
		break;
	case WM_LBUTTONDBLCLK:
		click_func(cfg_dblclickfunc);
		break;
	case WM_LBUTTONUP:
		click_func(cfg_lftclickfunc);
		break;
	case WM_MBUTTONUP:
		click_func(cfg_mdlclickfunc);
		break;
	case WM_PAINT:
		paint();
		break;
	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO mmi = LPMINMAXINFO(lp);
			mmi->ptMinTrackSize.y = cfg_minheight;
		}
		break;
	case WM_FOO_CALLBACK:
		switch(wp)
		{
		case PLAYBACK_NEW_TRACK:
			{
				metadb_handle * track = reinterpret_cast<metadb_handle*>(lp);
				set_handle(plytrack,track);
				if(!cfg_selected && !image_locked && set_image(plytrack))
					redraw();
			}
			break;
		case PLAYBACK_STOP:
			if(!image_locked && seltrack && set_image(seltrack))
				redraw();
			break;
		case PLAYLIST_FOCUS_CHANGE:
			{
				metadb_handle * track=playlist_oper::get()->get_item((int)lp);
				if(track){
					set_handle(seltrack,track);
					track->handle_release();
				}
			}
			if (cfg_selected || !play_control::get()->is_playing()){
				if(seltrack)
				{
					if(!image_locked && set_image(seltrack))
						redraw();
				}
			}
			break;
		case VARIABLE_CHANGE:
			{
				cfg_var * var = reinterpret_cast<cfg_var*>(lp);
				if(var && strcmp_partial(var->get_name(),"albumart_min_height")==0){
					host->on_size_limit_change(wnd1,ui_extension_flag::SLC_MIN_HEIGHT);
				}else if(var && strcmp_partial(var->get_name(),"albumart_use_control")==0){
					lock(false);
					if(set_image())
						redraw();
				}else if(var && strcmp_partial(var->get_name(),"albumart_srclist")==0){
					crnsource=0;
				}
			}
			if(!image_locked)	set_image();
			redraw();
			break;
		}
		break;
	}
	return uDefWindowProc(wnd1, msg, wp, lp);
}

void uie_albumart::build_menu(HMENU menu, unsigned base, POINT pt, bool show_shortcuts, void * & user_data)
{
	if (GetMenuItemCount(menu)){
		const unsigned id_offset=100;
		uAppendMenu(menu,(MF_SEPARATOR),0,"");
		for(int i=1;i<tabsize(func_name);i++)
			uAppendMenu(menu,(MF_STRING),base+id_offset+i,func_name[i]);
		CheckMenuItem(menu,base+100+FUNC_FOLLOW,cfg_selected?MF_CHECKED:MF_UNCHECKED);
		uAppendMenu(menu,(MF_SEPARATOR),0,"");
		uAppendMenu(menu,(MF_STRING),base,"Preferences...");
		user_data=(void*)id_offset;
	}
}

void uie_albumart::menu_action(HMENU menu, unsigned base, unsigned id, void * user_data)
{
	if (id == base)
	{
		play_control::get()->show_config("Album Art Panel");
	}else{
		click_func(id - (base + (unsigned)user_data));
	}
}

void uie_albumart::redraw()
{
	RedrawWindow(wnd, NULL, NULL, RDW_INVALIDATE | RDW_ERASE);
}

void uie_albumart::click_func(unsigned func)
{
	switch(func){
	case FUNC_OPENEX:
		func_OpenEx();
		break;
	case FUNC_OPENDIR:
		func_OpenDir();
		break;
	case FUNC_FOLLOW:
		func_ToggleFollow();
		break;
	case FUNC_FOCUSNP:
		func_FocusPlaying();
		break;
	case FUNC_REFRESH:
		func_Refresh();
		break;
	case FUNC_NEXTSRC:
		func_NextSource();
		break;
	case FUNC_RESETSRC:
		func_ResetSource();
		break;
	}
}

void uie_albumart::func_OpenEx()
{
	if(!img.is_empty())
		uShellExecute(0,"open",img,0,0,SW_SHOW);
}

void uie_albumart::func_OpenDir()
{
	if(crntrack){
		string8 temp;
		play_control::get()->playback_format_title_ex(
			crntrack,
			temp,
			"$replace(%_path%,%_filename_ext%,)",NULL,false,false);
		uShellExecute(0,"open",temp,0,0,SW_SHOW);
	}
}

void uie_albumart::func_ToggleFollow()
{
	cfg_selected=!cfg_selected;
}

void uie_albumart::func_FocusPlaying()
{
	playlist_oper::get()->set_focus(playlist_oper::get()->get_now_playing());
}

void uie_albumart::func_Refresh()
{
	if(!image_locked)	set_image();
	redraw();
}

void uie_albumart::func_NextSource()
{
	if(crntrack){
		unsigned n,m = cfg_src_list.get_count();
		n=crnsource+1;
		if(n>=m-1)
			n=0;
		for(;string8(cfg_src_list.get_value(n)).find_first("-")==0;n++){
			if(n==crnsource)
				break;
			else if(n>=m-1)
				n=-1;
		}
		crnsource=n;
		if(set_image(crntrack,crnsource))
			redraw();
	}
}

void uie_albumart::func_ResetSource()
{
	if(crntrack && set_image(crntrack,(crnsource=0)))
		redraw();
}

bool uie_albumart::set_image(){
	return set_image(!play_control::get()->is_playing()?seltrack:cfg_selected?seltrack:plytrack);
}

bool uie_albumart::set_image(unsigned fstsource){
	return set_image(!play_control::get()->is_playing()?seltrack:cfg_selected?seltrack:plytrack,fstsource);
}

bool uie_albumart::set_image(metadb_handle * track){
	if(!cfg_resetsource)
		return set_image(track,crnsource);
	else{
		string8 img_org=img;
		if(set_image(track,crnsource)){
			crnsource=0;
			img=img_org;
			return set_image(track,crnsource);
		}else
			return false;
	}
}

bool uie_albumart::set_image(metadb_handle * track, unsigned fstsource)
{
	unsigned n,m = cfg_src_list.get_count();
	if(!track || fstsource>=m) return false;
	string8 img_org=img;
	img.reset();
	set_handle(crntrack,track);
	if(m==0) return true;
	for(n=fstsource;n<m;n++)
	{
		string8 pattern=cfg_src_list.get_value(n);
		if(pattern.find_first("-")==0)
			pattern.remove_chars(0,1);
		bool match=false;
		if(pattern.find_first("match:")==0){
			match=true;
			pattern.remove_chars(0,6);
		}
		string8 file;
		play_control::get()->playback_format_title_ex(crntrack,file,pattern,NULL,false,false);
		if(match){
			file=findFirstImageMatch(file);
		}
		if(!file.is_empty()){
			Bitmap bmp(string_wide_from_utf8(file.get_ptr()));
			if(bmp.GetHeight()){
				img=file;
				break;
			}
		}
		if(n==fstsource-1 || (n==m-1 && fstsource==0)) break;
		if(n==m-1) n=-1;
	}
	return img.is_empty() ? !img_org.is_empty() : strcmp_partial(img_org,img)!=0 ;
}

void uie_albumart::paint()
{
	RECT wndrect_temp;
	if(GetClientRect(wnd,&wndrect_temp)){
		PAINTSTRUCT lpPaint;
		HDC hDC=BeginPaint(wnd,&lpPaint);
		Graphics gfx(hDC);
		Rect wndrect(
			wndrect_temp.left,
			wndrect_temp.top,
			wndrect_temp.right-wndrect_temp.left,
			wndrect_temp.bottom-wndrect_temp.top
			);
		FillRect(hDC,&wndrect_temp,(HBRUSH)(COLOR_BTNFACE+1));
		if(cfg_usebg){
			gfx.FillRectangle(
				& SolidBrush(Color((cfg_bgcol) & 0xFF,(cfg_bgcol>> 8) & 0xFF,(cfg_bgcol>> 16) & 0xFF)),
				wndrect
				);
		}
		if(!img.is_empty()){
			Bitmap bmp(string_wide_from_utf8(img.get_ptr()));
			if(cfg_padding>0){
				wndrect.Height-=2*cfg_padding;
				wndrect.Width-=2*cfg_padding;
				wndrect.X+=cfg_padding;
				wndrect.Y+=cfg_padding;
			}
			RectF rect(wndrect.X,wndrect.Y,bmp.GetWidth(),bmp.GetHeight());
			if(
				(cfg_expandalbum && rect.GetRight() < wndrect.GetRight()  &&  rect.GetBottom() < wndrect.GetBottom())  ||
				(cfg_shrinkalbum  &&  (rect.GetRight() > wndrect.GetRight()  ||  rect.GetBottom() > wndrect.GetBottom()))){
				if(!cfg_aspectratio){
					rect.Width=wndrect.GetRight();
					rect.Height=wndrect.GetBottom();
				}
				else if( (wndrect.GetRight()-rect.GetRight()) < (wndrect.GetBottom()-rect.GetBottom()) ){
					rect.Height*=((double)wndrect.GetRight()/rect.GetRight());
					rect.Width=wndrect.GetRight();
				}else{
					rect.Width*=((double)wndrect.GetBottom()/rect.GetBottom());
					rect.Height=wndrect.GetBottom();
				}
			}
			if(cfg_centeralbum)
				rect.Offset((double)(wndrect.GetRight()-rect.GetRight())/2,(double)(wndrect.GetBottom()-rect.GetBottom())/2);
			if(cfg_pixelborder){
				rect.Width--;
				rect.Height--;
				gfx.DrawRectangle(& Pen(Color(0,0,0)),rect);
				rect.X++;
				rect.Y++;
				rect.Width--;
				rect.Height--;
			}
			gfx.SetInterpolationMode(cfg_interpolationmode==2?InterpolationModeHighQuality:
					cfg_interpolationmode==1?InterpolationModeDefault:
					cfg_interpolationmode==0?InterpolationModeLowQuality:
					InterpolationModeHighQualityBicubic);
			int srcWidth=bmp.GetWidth(),srcHeight=bmp.GetHeight(),srcX=0,srcY=0;
			if(cfg_padding<0 && abs(cfg_padding)<bmp.GetWidth() && abs(cfg_padding)<bmp.GetHeight()){
				srcWidth+=2*cfg_padding;
				srcHeight+=2*cfg_padding;
				srcX-=cfg_padding;
				srcY-=cfg_padding;
			}
			gfx.DrawImage(
				&bmp,
				rect,  // destination rectangle 
				srcX, srcY,        // upper-left corner of source rectangle
				srcWidth,       // width of source rectangle
				srcHeight,      // height of source rectangle
				UnitPixel);
		}
		EndPaint(wnd,&lpPaint);
		gfx.ReleaseHDC(hDC);
	}
}

void uie_albumart::lock(bool l){
	image_locked=l;
}

bool uie_albumart::is_locked(){
	return image_locked;
}