#define UNICODE

#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"
#include "../shared/ui_extension_with_helpers.h"
#include <gdiplus.h>
using namespace Gdiplus;
#pragma comment(lib, "gdiplus.lib")
#include "resource.h"
#include "albumart_helpers.h"
#include "albumart_vars.h"
#include "albumart_notifiers.h"
#include "albumart_config.h"

class uie_albumart : public ui_extension_base_t<
	(ui_extension_flag::ui_extension_type)(ui_extension_flag::TYPE_PANEL|ui_extension_flag::TYPE_TOOLBAR), UEM_SINGLE >
{
protected:
	metadb_handle * seltrack;
	metadb_handle * plytrack;
	metadb_handle * crntrack;
	unsigned crnsource;
	string8 img;
	bool image_locked;

	GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;

public:
	virtual const GUID & get_extension_guid() const {
		// {E32B3859-B6BC-48e1-8370-3D5F5FA00A60}
		static const GUID guid = 
		{ 0xe32b3859, 0xb6bc, 0x48e1, { 0x83, 0x70, 0x3d, 0x5f, 0x5f, 0xa0, 0xa, 0x60 } };
		return guid;
	}
	virtual void get_name(string_base & out) const { out = "Album Art"; }
	virtual void get_category(string_base & out) const {out = "Panels";	}

	static LRESULT WINAPI host_proc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);
    LRESULT WINAPI on_message(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);
	
	virtual HWND create_window(HWND wnd_parent);

	virtual void build_menu(HMENU menu, unsigned base, POINT pt, bool show_shortcuts, void * & user_data);
	virtual void menu_action(HMENU menu, unsigned base, unsigned id, void * user_data);

	void paint();
	void redraw();
	void click_func(unsigned func);

	void func_OpenEx();
	void func_OpenDir();
	void func_ToggleFollow();
	void func_FocusPlaying();
	void func_Refresh();
	void func_NextSource();
	void func_ResetSource();

	bool set_image();
	bool set_image(unsigned fstsource);
	bool set_image(metadb_handle * track);
	bool set_image(metadb_handle * track,unsigned fstsource);

	void lock(bool l);
	bool is_locked();
};

static ui_extension_factory_single_transparent<uie_albumart> foo_uie_albumart;

class NOVTABLE albumart_control_base : public service_base
{
public:
	static GUID get_class_guid()
	{
		// {5549FA34-673C-4c5b-B6CE-5CDC4813E7B4}
		static const GUID guid = 
		{ 0x5549fa34, 0x673c, 0x4c5b, { 0xb6, 0xce, 0x5c, 0xdc, 0x48, 0x13, 0xe7, 0xb4 } };
		return(guid);
	}

	static albumart_control_base* get() 
	{
		return(service_enum_create_t(albumart_control_base,0));
	}
	
	virtual bool is_controllable()=0;
	virtual bool set_image(metadb_handle * track)=0;
	virtual bool set_image()=0;
	virtual void redraw()=0;
	virtual void lock(bool l)=0;
	virtual bool is_locked()=0;
};

class albumart_control : public albumart_control_base
{
	virtual bool is_controllable(){
		return cfg_usecontrol && foo_uie_albumart.get_wnd();
	}
	virtual bool set_image(metadb_handle * track)
	{
		if(cfg_usecontrol && foo_uie_albumart.get_wnd())
			return foo_uie_albumart.set_image(track);
		else
			return NULL;
	}
	virtual bool set_image()
	{
		if(cfg_usecontrol && foo_uie_albumart.get_wnd())
			return foo_uie_albumart.set_image();
		else
			return NULL;
	}
	virtual void redraw()
	{
		if(cfg_usecontrol && foo_uie_albumart.get_wnd())
			foo_uie_albumart.redraw();
	}
	virtual void lock(bool l)
	{
		if(cfg_usecontrol && foo_uie_albumart.get_wnd())
			foo_uie_albumart.lock(l);
	}
	virtual bool is_locked()
	{
		if(cfg_usecontrol && foo_uie_albumart.get_wnd())
			return foo_uie_albumart.is_locked();
		else
			return NULL;
	}
};

static service_factory_single_transparent_t<albumart_control_base,albumart_control> foo_uie_albumart_control;